-- All code written and maintained by Yewchi 
-- zyewchi@gmail.com

-- Rushing to get this into a working Alpha. Most code is functional and 
--    basically extendable, so turning full modular will be easy later on, 
--    after some abstraction.

--- Initialization --
--
------- Filenames --
--
AUDIOQS_ROOT = 		"Interface/AddOns/AudioQs/"
SOUNDS_ROOT = 		AUDIOQS_ROOT.."Sounds/"
--
------ /Filenames --

------- Flags --
--
PROMPT_STAGE_OFF = 									0x0

PROMPTSEG_CONDITIONAL_USE_PREVIOUS =				0xFFF0
PROMPTSEG_CONDITIONAL_USE_STOP =					0xFFF1
PROMPTSEG_CONDITIONAL_CONTINUATION =				0xFFF2
PROMPTSEG_CONDITIONAL_REPEATER =					0xFFF3

SPELL_TYPE_ABILITY = 								0xFFF0
SPELL_TYPE_AURA = 									0xFFF1

SPEC_NOT_IMPLEMENTED =								0xFFFF
--
------ /Flags --

------- Table key references --
--
SPELL_SPELL_NAME = 1
SPELL_CHARGES = 2
SPELL_DURATION = 3
SPELL_EXPIRATION = 4
SPELL_UNIT_ID = 5
SPELL_SPELL_TYPE = 6
SPELL_PROMPT_INDEX = 7 -- TODO Defunct

SEGT_CONDITIONALS =	1
SEGT_FIRST = 2
SEGT_CONDITIONALS_START = 1
SEGT_CONDITIONALS_STOP = 2

PROMPT_TIMESTAMP = 1
PROMPT_CURR_STAGE = 2
PROMPT_SPELL_ID = 3
PROMPT_SEGMENT_TABLE_INDEX = 4

PROMPTSEG_LENGTH = 1
PROMPTSEG_FILE = 2
PROMPTSEG_HANDLE = 3
PROMPTSEG_CONDITIONAL = 4

SPELL_INFO_NAME = 1
SPELL_INFO_ID = 7

UNIT_AURA_NAME = 1
UNIT_AURA_COUNT = 3
UNIT_AURA_DURATION = 5
UNIT_AURA_EXPIRATION = 6
UNIT_AURA_SPELL_ID = 10
UNIT_AURA_TIME_MOD = 15 -- always = 1. For prelim confirmation of a valid aura

COMBAT_LOG_SOURCE_GUID = 4
COMBAT_LOG_SPELL_ID = 12
--
------ /Table key references --

------- Static vals --
--
DEBUG = false
VERBOSE = DEBUG and true
local MAX_ERRORS_PER_SESSION = 3
local DUPLICATE_ERROR_GARBAGE_TIMER = 3.0

AURA_ALWAYS_IS_VALUE = 1 -- using value #15: timeMod

PLAYER_GUID = UnitGUID("player")

AUDIO_CHANNEL = 									"DIALOG"

-- Auto-generate
local SPECS = {
		["Beast Mastery"] = 253
}
--
------- /Static vals --

------- Return values --
--
SPELLS_SPELL_NOT_LISTED = 							0xFFF0
--
------ /Return values --

------- Error Codes --
--
ERR_UNKNOWN =										0xEFF0
ERR_INVALID_ARGS = 									0xEFF1
ERR_UNKNOWN_SPELL_AS_ARGUMENT = 				0xEFF2
ERR_INVALID_AURA_DATA =								0xEFF3
--
------ /Error Codes --

------- Default Msgs --
--
audioQsSpecifier = "|cff50C0F0[AudioQs]".."|r"
debugSpecifier = "|cffF080F0<debug>".."|r: "
errSpecifier = "|cffF08050<error>".."|r: "
ERR_MSGS = { 
	[ERR_UNKNOWN] =
		audioQsSpecifier..
		errSpecifier..
		"#"..ERR_UNKNOWN.." "..
		"Unknown error.",
	[ERR_INVALID_ARGS] =
		audioQsSpecifier..
		errSpecifier..
		"#"..ERR_INVALID_ARGS.." "..
		"Invalid args.",
	[ERR_UNKNOWN_SPELL_AS_ARGUMENT] = 
		audioQsSpecifier..
		errSpecifier..
		"#"..ERR_UNKNOWN_SPELL_AS_ARGUMENT.." "..
		" An unlisted spellId was passed to a function.",
	[ERR_INVALID_AURA_DATA] = 
		audioQsSpecifier..
		errSpecifier..
		"#"..ERR_INVALID_AURA_DATA.." "..
		" Data passed to a function did not contain a valid aura table."
}
STOP_ERROR_MAX_REPORTS = audioQsSpecifier.." Max errors exceeded. Stopping error reports."
--
------ /Default msgs --

------- AddOn variables --
--
local errorsThisSession = 0
local mostRecentErrorString = nil
local mostRecentErrorTimestamp = 0

local specId = nil
--
------ /AddOn variables --
--
-- /Initialization --

--- Funcs --
--
-------------- HandleError()
local function HandleError(err, parent, child)
	if not DEBUG and errorsThisSession >= MAX_ERRORS_PER_SESSION then 
		return
	end
	
	if type(err) == "table" then 
		if err.code == nil then err.code = ERR_UNKNOWN end
	elseif err == nil then 
		err = {code=ERR_UNKNOWN}
	end
	
	-- Formulate full string --
	local msgGiven = (type(err) == "string" and err or "")
	local errorString = ""
	local funcString = (err.func~=nil and " func:"..err.func or (parent~=nil and parent or "???").."->"..(child~=nil and child or "???"))
	
	if type(err.code) == "number" then 
		local errMsg = ERR_MSGS[err.code]
		errorString = (errMsg~=nil and errMsg or "[InvalidErrCode]")..funcString
	else
		errorString = ERR_MSGS[ERR_UNKNOWN].." "..msgGiven.." "..funcString
	end
	
	-- Check error frequency of duplicates is not too high, and print. --
	local currTime = GetTime()
	if not (errorString == mostRecentErrorString and
			currTime < mostRecentErrorTimestamp + DUPLICATE_ERROR_GARBAGE_TIMER) then
		mostRecentErrorString = errorString
		mostRecentErrorTimestamp = currTime
		
		print(errorString)
		
		errorsThisSession = errorsThisSession + 1
		if not DEBUG and errorsThisSession >= MAX_ERRORS_PER_SESSION then
			print(STOP_ERROR_MAX_REPORTS)
		end
	end
end

-------------- GetSpellId()
local function GetSpellId(s)
	local spellId = spellsLookUp[s]
	return (spellId ~= nil and spellId or SPELLS_SPELL_NOT_LISTED)
end

-------------- FindPromptsFromUnitAura()
local function FindPromptsFromUnitAura(aura)
	if aura[UNIT_AURA_TIME_MOD] ~= AURA_ALWAYS_IS_VALUE then
		error({code=ERR_INVALID_AURA_DATA, func="FindPromptsFromUnitAura() aura=", unpack(aura)})
	end
	
	local spellId = aura[UNIT_AURA_SPELL_ID]
	if AuraIsIncluded(spellId) then 
		return ProcessAuraForPrompts(unpack(aura))
	end
end

-------- GetAuraInfo()
function GetAuraInfo(unitId, spellId, spellType)
	for n=1,40,1 do
		local auraSpellId = select(UNIT_AURA_SPELL_ID, UnitAura(unitId, n, spellType))
		if auraSpellId == nil then 
			return nil
		elseif auraSpellId == spellId then
			return UnitAura(unitId, n, spellType)
		end
	end
end

-- Needs confirmation, error thrown on missing file
-------- PlaySoundGetHandle()
function PlaySoundGetHandle(soundFile)
if VERBOSE then print("Playing \""..soundFile.."\"") end
	return select(2, PlaySoundFile(soundFile, AUDIO_CHANNEL))
end
--
-- /Funcs --

--- Event Funcs --
--
local Frame_SpellUpdateCooldown = CreateFrame("Frame", "Spell Update Charges")
local Frame_CombatLog = CreateFrame("Frame", "Combat Log")
local Frame_UnitAura = CreateFrame("Frame", "Unit Aura")
local Frame_LoadOrSpecChange = CreateFrame("Frame", "Load or Spec Change")

-------------- RegisterEvents()
local function RegisterEvents()
	Frame_SpellUpdateCooldown:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	Frame_SpellUpdateCooldown:RegisterEvent("SPELL_UPDATE_CHARGES")
	Frame_CombatLog:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	Frame_UnitAura:RegisterEvent("UNIT_AURA")
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Prompt tracking enabled.") end
end

-------------- UnregisterEvents()
local function UnregisterEvents()
	Frame_CombatLog:UnregisterAllEvents()
	Frame_SpellUpdateCooldown:UnregisterAllEvents()
	Frame_UnitAura:UnregisterAllEvents()
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Prompt tracking disabled.") end
end

-------------- LoadSpec()
local function UpdateSpecializationInfo(self, event, ...)
	newSpecId = GetSpecializationInfo(GetSpecialization())
	local eventsWereRegistered = specId ~= nil and SpecHasPrompts(specId)
	
	specId = newSpecId
	
	local success, err = pcall(LoadSpecialization)
	if not success then HandleError(err, "UpdateSpecializationInfo()", "LoadSpecialization()") end -- More robust coverage
	
	local specHasPrompts = SpecHasPrompts(newSpecId)
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Spec #"..newSpecId.." has "..(specHasPrompts == true and "" or "no ").."available prompts.") end
	if specHasPrompts then
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Loaded spec "..specId) end
		if not eventsWereRegistered then -- TODO As event prompts are added, requires an autmatically generated list of frames/events to call for the rules. Event registration needs a more robust process
			RegisterEvents()
		end
	elseif EventsWereRegistered then
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Turning off AudioQs") end
		UnregisterEvents()
	end
end	
Frame_LoadOrSpecChange:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
Frame_LoadOrSpecChange:RegisterEvent("PLAYER_LOGIN")
Frame_LoadOrSpecChange:SetScript("OnEvent", function(self, event, ...) UpdateSpecializationInfo(self, event, ...) end)

Frame_CombatLog:SetScript("OnEvent",
	function(self, event, ...)
		local success, err = pcall(ProcessCombatLogForPrompts)
		if not success then HandleError(err, "CombatLogOnEvent", "ProcessCombatLogForPrompts()") end
	end
)

Frame_SpellUpdateCooldown:SetScript("OnEvent",
	function(self, event, ...)
		local success, err = pcall(ProcessSpellCooldownsForPrompts)
		if not success then HandleError(err, "SpellUpdateCooldownOnEvent()", "ProcessSpellCooldownsForPrompts()") end
	end
)

-- Optimize for already completed checks, no further spell prompts possible, etc.
Frame_UnitAura:SetScript("OnEvent",
	function(self, event, ...)
		local unitId = ...
		
		if not UnitIsIncluded(unitId) then
			return
		end
		
		if specId ~= nil then
		-- Abstract into class file funcs
			for n = 1, 40, 1 do
				local aura = {UnitAura(unitId, n, "HELPFUL")}
				if aura[UNIT_AURA_SPELL_ID] ~= nil then 
					local success, err = pcall(FindPromptsFromUnitAura, aura)
					if not success then HandleError(err, "UnitAuraOnEvent()", "FindPromptsFromUnitAura()") break end
				else
					break 
				end
			end
		end
	end
)
--
-- /Event Funcs --