-- All code written and maintained by Yewchi
-- zyewchi@gmail.com

-- Needs further abstraction. Could still add Prompt.lua, to separate segment functions, even the ticker, this could be simply a loading and table generation module.

--- Initialization --
------- AddOn variables --
--
spells = {}
spellsSnapshot = {}
spellsLookUp = {}
local abilityTable = {}
local auraTable = {}

local segments = {}

local promptsTable = {}
local segmentPromptIndex = {}

local unitsIncluded = {}
--
------ /AddOn variables --


Frame_PromptTicker = CreateFrame("Frame", "Prompt Ticker")
--
-- /Initialization --



--- Funcs --
--
local CheckPrompts
local UpdateSpellTable
local AttemptStartPrompts
local PromptTicker

-------------- EvaluateConditional()
local function EvaluateConditional(conditional)
	if type(conditional) == "string" then 
		return loadstring(conditional)()
	end
	return conditional == true
end

-------------- StopAllSegmentsSounds()
local function StopAllSegmentsSounds(segmentsArray)
	for n=SEGT_FIRST,#segmentsArray,1 do
		local soundHandle = segmentsArray[n][PROMPTSEG_HANDLE]
		if soundHandle ~= nil then
			StopSound(soundHandle)
			segmentsArray[n][PROMPTSEG_HANDLE] = nil
		end
	end
end

-------------- CheckStopSegments()
local function CheckStopSegments(segmentsArray, prompt, allowContinuationOfPlay, forceStop)
	if EvaluateConditional(segmentsArray[SEGT_CONDITIONALS][SEGT_CONDITIONALS_STOP]) or forceStop == true then
		if allowContinuationOfPlay == false then
			StopAllSegmentsSounds(segmentsArray)
		end
		prompt[PROMPT_TIMESTAMP] = 0
		prompt[PROMPT_CURR_STAGE] = PROMPT_STAGE_OFF
		return true
	end
	return false
end

-------------- GetNextTrueSegment()
local function GetNextTrueSegment(segmentsArray, nextPromptStage)
	if nextPromptStage > #segmentsArray then
		return PROMPT_STAGE_OFF
	end
	
	local skippingFalseConditionalSet = false
	for n=nextPromptStage,#segmentsArray,1 do -- Search downards for true conditionals
		local conditional = segmentsArray[n][PROMPTSEG_CONDITIONAL]
		if conditional == PROMPTSEG_CONDITIONAL_USE_PREVIOUS then
			if skippingFalseConditionalSet == false then
				for p=n-1,SEGT_FIRST,-1 do -- search upwards for this conditional set's validity
					local conditional = segmentsArray[p][PROMPTSEG_CONDITIONAL]
					if conditional == nil then 
						return SEGT_STOP -- TODO if p ~= SEGT_FIRST then error(SEGT_MALFORMED_SEGMENT_TABLE) ?
					end
					
					if conditional ~= PROMPTSEG_CONDITIONAL_USE_PREVIOUS then
						if EvaluateConditional(conditional) then
							return n
						end
						skippingFalseConditionalSet = true
						break
					end
				end
			end
		elseif conditional == PROMPTSEG_CONDITIONAL_USE_STOP then 
			if EvaluateConditional(segmentsArray[SEGT_CONDITIONALS][SEGT_CONDITIONALS_STOP]) then
				return n
			end
			skippingFalseConditionalSet = true
		elseif conditional == PROMPTSEG_CONDITIONAL_CONTINUATION and skippingFalseConditionalSet == false then
			return n
		else
			if EvaluateConditional(conditional) then 
				return n
			end
			skippingFalseConditionalSet = true
		end
		if segmentsArray[n][PROMPTSEG_LENGTH] == nil and type(segmentsArray[n+1]) ~= "table" then
			break
		end
	end
	return PROMPT_STAGE_OFF
end

----- Continues the running of PromptTicker if any prompts are in-progress.
-------------- CheckPrompts()
CheckPrompts = function()
	for n=#promptsTable,1,-1 do
		if promptsTable[n][PROMPT_CURR_STAGE] ~= PROMPT_STAGE_OFF then
			if Frame_PromptTicker:GetScript("OnUpdate") == nil then
				Frame_PromptTicker:SetScript("OnUpdate", function() PromptTicker() end)
			end
			return
		end
	end
	if Frame_PromptTicker:GetScript("OnUpdate") ~= nil then
		Frame_PromptTicker:SetScript("OnUpdate", nil)
	end
end

-------------- AttemptStartPrompt()
AttemptStartPrompt = function(spellId, runCheckPrompts)
	local segmentPromptsArray = segmentPromptIndex[spellId]
	if segmentPromptsArray == nil then
		error({code=ERR_UNKNOWN_SPELL_AS_ARGUMENT, func="AttemptStartPrompt() spellId="..(spellId~=nil and spellId or "nil").." runCheckPrompts="..(runCheckPrompts~=nil and runCheckPrompts or "nil")})
	end
	
	local spellSegmentsTable = segments[spellId]
	for n=1,#segmentPromptsArray,1 do -- For each set of segments pertaining to this spellId
		local thisSegments = spellSegmentsTable[n]
		local thisSegmentsPrompt = promptsTable[segmentPromptsArray[n]]
		local startingSegment = GetNextTrueSegment(thisSegments, SEGT_FIRST)

		if EvaluateConditional(thisSegments[SEGT_CONDITIONALS][SEGT_CONDITIONALS_START]) then
			CheckStopSegments(thisSegments, thisSegmentsPrompt, false, true) -- TODO--Design decis. Always stop if the spell has been updated?
			
			thisSegmentsPrompt[PROMPT_TIMESTAMP] = GetTime()
			thisSegmentsPrompt[PROMPT_CURR_STAGE] = startingSegment
			
			if thisSegments[startingSegment][PROMPTSEG_FILE] ~= nil then 
				thisSegments[startingSegment][PROMPTSEG_HANDLE] = PlaySoundGetHandle(thisSegments[startingSegment][PROMPTSEG_FILE])
			end
		end
		CheckStopSegments(thisSegments, thisSegmentsPrompt, false)
		if runCheckPrompts ~= false then 
			CheckPrompts()
		end
	end
end

-------------- SaveSpellsSnapshot()
local function SaveSpellsSnapshot(spellId)
	if spellsSnapshot[spellId] == nil then spellsSnapshot[spellId] = {} end
	
	for n=1,#spells[spellId],1 do
		spellsSnapshot[spellId][n] = spells[spellId][n]
	end
end

-------------- UpdateAllSpellInfo()
local function UpdateAllSpellInfo()
	for n=1,#promptsTable,1 do
		local spellId = promptsTable[n][PROMPT_SPELL_ID]
		if promptsTable[n][PROMPT_SEGMENT_TABLE_INDEX] == 1 then
			local cdStart, cdDur = GetSpellCooldown(spellId)
			local cdExpiration = cdStart + cdDur
			
			UpdateSpellTable(spellId, cdDur, cdExpiration)
		end
	end
end

-------------- GenerateTablesAndIndices()
local function GenerateTablesAndIndices()
	do
		local i = 1
		local j = 1
		for k,v in pairs(spells) do
			-- spellsLookUp
			spellsLookUp[v[SPELL_SPELL_NAME]] = k
			-- abilityTable
			local previousAbilitySpellId = (#abilityTable > 0 and abilityTable[i-1] or nil)
			if v[SPELL_SPELL_TYPE] == SPELL_TYPE_ABILITY and previousAbilitySpellId ~= k then
				abilityTable[i] = k
				i = i + 1
			end
			-- auraTable
			local previousAuraSpellId = (#auraTable > 0 and auraTable[j-1] or nil)
			if v[SPELL_SPELL_TYPE] == SPELL_TYPE_AURA and previousAuraSpellId ~= k then
				auraTable[j] = k
				j = j + 1
			end
		end
	end
	
	do
		local i = 1
		for k,v in pairs(segments) do
			segmentPromptIndex[k] = {}
			for n=1,#v,1 do
				-- promptsTable
				promptsTable[i] = {	0, PROMPT_STAGE_OFF, k, n } -- PROMPT_TIMESTAMP, PROMPT_CURR_STAGE, PROMPT_SPELL_ID, PROMPT_SEGMENT_TABLE_INDEX
				-- segmentPromptIndex
				segmentPromptIndex[k][n] = i
				i = i + 1
			end
		end
	end
	-- unitsIncluded
	for n=1,#promptsTable,1 do
		local thisSpellId = promptsTable[n][PROMPT_SPELL_ID]
		local thisSegmentIndex = promptsTable[n][PROMPT_SEGMENT_TABLE_INDEX]
		if thisSegmentIndex == 1 then
			local thisUnit = spells[thisSpellId][SPELL_UNIT_ID]
			if thisUnit ~= nil then
				if #unitsIncluded == 0 then
					unitsIncluded[#unitsIncluded+1] = thisUnit
				else
					local insert = true
					for n=1,#unitsIncluded,1 do
						if unitsIncluded[n] == thisUnit then
							insert = false
							break
						end
					end
					if insert then
						unitsIncluded[#unitsIncluded+1] = thisUnit
					end
				end
			end
		end
	end
end

-------------- UpdateSpellTable()
UpdateSpellTable = function(spellId, cdDur, cdExpiration)
	if spells[spellId] == nil then
		error({code=ERR_UNKNOWN_SPELL_AS_ARGUMENT, func="UpdateSpellTable() spellId="..(spellId~=nil and spellId or "nil").." cdDur="..(cdDur~=nil and cdDur or "nil").." cdExpiration="..(cdExpiration~=nil and cdExpiration or "nil")})
	end
	local thisSpell = spells[spellId]
	
	if cdDur == nil or cdExpiration == nil then
if DEBUG then if cdDur~=cdExpiration then print(audioQsSpecifier..debugSpecifier.."UpdateSpellTable(): - "..(cdDur == nil and "cdDur" or "cdExpiration").." was sole nil passed.") end end
		local start, dur = GetSpellCooldown(spellId)
		cdDur = dur
		cdExpiration = start + dur
	end
	
	if (thisSpell[SPELL_SPELL_TYPE] == SPELL_TYPE_AURA and thisSpell[SPELL_EXPIRATION] ~= cdExpiration) or
			(thisSpell[SPELL_SPELL_TYPE] == SPELL_TYPE_ABILITY and (thisSpell[SPELL_CHARGES] ~= GetSpellCharges(spellId) or thisSpell[SPELL_EXPIRATION] ~= cdExpiration)) then
		SaveSpellsSnapshot(spellId)
		thisSpell[SPELL_CHARGES] = GetSpellCharges(spellId)
		thisSpell[SPELL_DURATION] = cdDur
		thisSpell[SPELL_EXPIRATION] = cdExpiration
				
		return true
	end
	return false
end

-------- TGetFuncs()
function TGetFuncs()
	return SPEC_FUNCS
end

-------- TGetSpells()
function TGetSpells()
	return spells
end

-------- TGetSpellsLookup()
function TGetSpellsLookup()
	return spellsLookUp
end

-------- UnitIsIncluded()
function UnitIsIncluded(unitId)
	for n=1,#unitsIncluded,1 do
		if unitsIncluded[n] == unitId then
			return true
		end
	end
	return false
end

-------- AuraIsIncluded()
function AuraIsIncluded(spellId)
	return false ~= (spells ~= nil and spells[spellId] and spells[spellId][SPELL_SPELL_TYPE] == SPELL_TYPE_AURA or false)
end

-------- ProcessCombatLogForPrompts()
function ProcessCombatLogForPrompts()
	local combatLogEventInfo = {CombatLogGetCurrentEventInfo()}
	local spellId = combatLogEventInfo[COMBAT_LOG_SPELL_ID]
	local spell = spells[spellId]
	
	if spell == nil or combatLogEventInfo[COMBAT_LOG_SOURCE_GUID] ~= PLAYER_GUID then
		return
	end
	
	if spell[SPELL_SPELL_TYPE] == SPELL_TYPE_ABILITY then
		local cdStart, cdDur = GetSpellCooldown(spellId)
		local cdExpiration = cdStart + cdDur
		
		if UpdateSpellTable(spellId, cdDur, cdExpiration) then
			AttemptStartPrompt(spellId)
		else
			CheckPrompts()
		end
	end
end

-------- ProcessSpellCooldownForPrompts()
function ProcessSpellCooldownsForPrompts()
	for n=1,#abilityTable,1 do
		local spellId = abilityTable[n]
		local cdStart, cdDur = GetSpellCooldown(spellId)
		local cdExpiration = cdStart + cdDur
		
		if UpdateSpellTable(spellId, cdDur, cdExpiration) then
			AttemptStartPrompt(spellId)
		else
			CheckPrompts()
		end
	end
end

-- Puting full aura table processing in here, then call this func, such that
--   we don't require passing table spell refs to core.
-------- ProccessAuraForPrompts()
function ProcessAuraForPrompts(...)
	if ... == nil then
		error({code=ERR_INVALID_ARGS, func="ProcessAuraForPrompts() "..(...~=nil and ... or "nil")})
	end

	local name, _, count, _, cdDur, cdExpiration, _, _, _, spellId = ...
	local spellData = spells[spellId]
	
	if spellData == nil then 
		error({code=ERR_UNKNOWN_SPELL_AS_ARGUMENT, func="ProcessAuraForPrompts() "..(...~=nil and ... or "nil")})
	elseif UpdateSpellTable(spellId, cdDur, cdExpiration) then
		AttemptStartPrompt(spellId)
	end
end

-------- SpecHasPrompts()
function SpecHasPrompts(specId)
	return false ~= (SV_Specializations ~= nil and SV_Specializations[specId] or false)
end

-------- LoadSpecialization()
function LoadSpecialization()
	local newSpec = GetSpecializationInfo(GetSpecialization())
	
	if SV_Specializations == nil then
if DEBUG then print("Loading example spec.") end
		SV_Specializations = {}
		SV_Specializations[GetExampleSpec()] = GetExample()
	end
	
	-- TODO In-game edits will require save to SV_Specializations[previousSpec]
	
	if SV_Specializations[newSpec] ~= nil then
		spells = SV_Specializations[newSpec].spells
		events = SV_Specializations[newSpec].events
		segments = SV_Specializations[newSpec].segments
	end

	GenerateTablesAndIndices()
	UpdateAllSpellInfo()
if DEBUG then print(audioQsSpecifier..debugSpecifier.."Prompt tables generated. Spell info updated.") end
end
--
-- /Funcs --

--- Event Funcs --
--
-- Should register prompts instead of iteration over promptsTable. Regardless, CPU utilization still low.
-------------- PromptTicker()
PromptTicker = function()
	local currTime = GetTime()
	
	for n=#promptsTable,1,-1 do
		local promptTimestamp, promptStage, spellId, segmentsArrayIndex = unpack(promptsTable[n])
		if promptStage ~= PROMPT_STAGE_OFF then
			local thisSegments = segments[spellId][segmentsArrayIndex]
			if not CheckStopSegments(thisSegments, promptsTable[n]) then
				if currTime > promptTimestamp + thisSegments[promptStage][PROMPTSEG_LENGTH] then
					local nextPromptStage = GetNextTrueSegment(thisSegments, promptStage+1)
					local stopSegments = nextPromptStage == PROMPT_STAGE_OFF
					
					if not stopSegments then
						thisSegments[nextPromptStage][PROMPTSEG_HANDLE] = PlaySoundGetHandle(thisSegments[nextPromptStage][PROMPTSEG_FILE])
						promptsTable[n][PROMPT_TIMESTAMP] = currTime
						
						if thisSegments[nextPromptStage][PROMPTSEG_LENGTH] ~= nil then
							promptsTable[n][PROMPT_CURR_STAGE] = nextPromptStage
						else
							stopSegments = true
						end
					end
					
					if stopSegments then
						CheckStopSegments(thisSegments, promptsTable[n], true, true)
						CheckPrompts()
					end
				end
			end
		end
	end
end
--
-- /Event Funcs